/*** Z80Em: Portable Z80 emulator *******************************************/
/***                                                                      ***/
/***                              Z80Codes.h                              ***/
/***                                                                      ***/
/*** This file contains various macros used by the emulation engine       ***/
/***                                                                      ***/
/*** Copyright (C) Marcel de Kogel 1996,1997                              ***/
/***     You are not allowed to distribute this software commercially     ***/
/***     Please, notify me, if you make any changes to this file          ***/
/****************************************************************************/

#ifndef TRUE
#define TRUE (0==0)
#define FALSE (0!=0)
#endif

char *RS, *RS2, *Q; /* Used to avoid multiple macro expansion side-effects */

extern int pass;
int lab[0x10000];
char *tags[0x10000];

int deflab(int pc)
{
  if (pass == 2) return pc;
//  fprintf(stderr, "LABEL AT %04x\n", pc);
  lab[pc] = TRUE;
  return pc;
}

//#define M_POP(Rg)           \
//        R.Rg.D=M_RDSTACK(SP)+(M_RDSTACK((SP+1)&65535)<<8); \
//        SP+=2
char *M_POP_S(char *Rg)
{
  return S("%s=mem[SP++];%s+=(mem[SP++]<<8)", Rg, Rg);
}
#define M_POP(Rg) M_POP_S(#Rg)


char *M_POP2X_S(char *Rg, char *R1, char *R2)
{
  return S("%s=%s=mem[SP++];%s+=((%s=mem[SP++])<<8)", Rg, R2, Rg, R1);
}

#define M_POP2X(Rg,R1,R2) M_POP2X_S(#Rg,#R1,#R2)


char *M_POP2_S(char *R1, char *R2)
{
  return S("%s=mem[SP++];%s=mem[SP++]<<8 /* check order! */", R2, R1);
}
#define M_POP2(R1,R2) M_POP2_S(#R1,#R2)

void M_PUSH_S(char *Rg)
{
        D("  mem[--SP]=(%s)>>8; mem[--SP]=(%s)&255;", Rg, Rg);
}
#define M_PUSH(Rg) M_PUSH_S(#Rg)

void M_PUSH2_S(char *R1, char *R2)
{
        D("  mem[--SP]=%s; mem[--SP]=%s; /* check order */", R1, R2);
}
#define M_PUSH2(R1,R2) M_PUSH2_S(#R1,#R2)

char *M_CALL_S(void)
{
 int q=M_RDMEM_OPCODE_WORD();
 char *c;
 deflab(PC); deflab(q);
 c = S("mem[--SP]=0x%02x; mem[--SP]=0x%02x; goto L%04x", PC>>8, PC&255, q);
 PC=q;
 return c;
}
#define M_CALL M_CALL_S()

char *M_JP_S(void)
{
  deflab(PC+2);
// bug fix?  PC=M_RDOP_ARG(PC)+((M_RDOP_ARG((PC+1)&65535))<<8);
  PC=(M_RDOP_ARG(PC)&255) | ((M_RDOP_ARG(PC+1)&255)<<8);
  return S("goto L%04x", deflab(PC));
}

#define M_JP            M_JP_S()

char *M_JR_S(void)
{
  deflab(PC+1);// +1 is to skip the relative offset byte.  Then points to next inst
  PC+=((signed char)M_RDOP_ARG(PC))+1;
  return S("goto L%04x", deflab(PC));
}
#define M_JR            M_JR_S()

#define M_RET           "PC=mem[SP++]; PC+=(mem[SP++]<<8); goto *lab[PC]"

// now in z80.c #define M_RST(Addr)     M_PUSH(PC); PC=Addr
#define M_SET_S(Bit,Reg)  S("%s|=1<<%s", Reg, Bit)
#define M_SET(Bit,Reg)  M_SET_S(#Bit, #Reg)
#define M_RES(Bit,Reg)  S("%s&=~(1<<%s)", #Reg, #Bit)
#define M_BIT_S(Bit,Reg) S("RF=(RF&C_FLAG)|H_FLAG|((%s&(1<<%s))? ((%s==7)?S_FLAG:0):Z_FLAG)", Reg, RS=Bit, RS)
#define M_BIT(Bit,Reg)  M_BIT_S(#Bit, #Reg)
#define M_AND_S(Reg)    S("RA&=%s; RF=ZSPTable[RA]|H_FLAG /* ZSPHTable? */", Reg)
#define M_AND(Reg)      M_AND_S(#Reg)
#define M_OR_S(Reg)     S("RA|=%s; RF=ZSPTable[RA]", Reg)
#define M_OR(Reg)       M_OR_S(#Reg)
#define M_XOR_S(Reg)    S("RA^=%s; RF=ZSPTable[RA]", Reg)
#define M_XOR(Reg)      M_XOR_S(#Reg)
#define M_IN(Reg)       S("%s=DoIn(RC,RB); RF=(RF&C_FLAG)|ZSPTable[%s]", RS=#Reg, RS)

#define M_RLCA  "RA=(RA<<1)|((RA&0x80)>>7);\n  RF=(RF&0xEC)|(RA&C_FLAG)"

#define M_RRCA  "RF=(RF&0xEC)|(RA&0x01);\n  RA=(RA>>1)|(RA<<7)"

#define M_RLA   "i=RF&C_FLAG; RF=(RF&0xEC)|((RA&0x80)>>7); RA=(RA<<1)|i"

#define M_RRA   "i=RF&C_FLAG; RF=(RF&0xEC)|(RA&0x01); RA=(RA>>1)|(i<<7)"

#define M_RLC_S(Reg) (RS=Reg, Q=S("i=%s>>7; %s=(%s<<1)|i; RF=ZSPTable[%s]|i", RS, RS, RS, RS), Q)
#define M_RLC(Reg)   M_RLC_S(#Reg)

#define M_RRC_S(Reg) (RS=Reg, Q=S("i=%s&1; %s=(%s>>1)|(i<<7); RF=ZSPTable[%s]|i", RS, RS, RS, RS), Q)
#define M_RRC(Reg)   M_RRC_S(#Reg)

#define M_RL_S(Reg)  (RS=Reg, Q=S("i=%s>>7; %s=(%s<<1)|(RF&1); RF=ZSPTable[%s]|i;", RS, RS, RS, RS), Q)
#define M_RL(Reg)    M_RL_S(#Reg)

#define M_RR_S(Reg)  (RS=Reg, S("i=%s&1; %s=(%s>>1)|(RF<<7); RF=ZSPTable[%s]|i", RS, RS, RS, RS), Q)
#define M_RR(Reg)    M_RR_S(#Reg)

#define M_SLL_S(Reg) (RS=Reg, Q=S("i=%s>>7; %s=(%s<<1)|1; RF=ZSPTable[%s]|i", RS, RS, RS, RS), Q)
#define M_SLL(Reg)   M_SLL_S(#Reg)

#define M_SLA_S(Reg) (RS=Reg, Q=S("i=%s>>7; %s<<=1; RF=ZSPTable[%s]|i", RS, RS, RS),Q)
#define M_SLA(Reg)   M_SLA_S(#Reg)

#define M_SRL_S(Reg) (RS=Reg, Q=S("i=%s&1; %s>>=1; RF=ZSPTable[%s]|i", RS, RS, RS), Q)
#define M_SRL(Reg)   M_SRL_S(#Reg)

#define M_SRA_S(Reg) (RS=Reg, Q=S("i=%s&1; %s=(%s>>1)|(%s&0x80); RF=ZSPTable[%s]|i", RS=Reg, RS, RS, RS, RS), Q)
#define M_SRA(Reg)   M_SRA_S(#Reg)

// |ZSTable[RB]|((RB==0x80)?V_FLAG:0)|((RB&0x0F)?0:H_FLAG)
// is now replaced with |IncTable[RB]

#define M_INC(Reg) "RF=(RF&C_FLAG)|IncTable[++" #Reg "]"

#define M_DEC(Reg) S("RF=DecTable[--%s]", #Reg)
#define M_DEC_S(Str) S("RF=DecTable[--%s]", Str)

#define M_ADD_S(Reg) (RS=Reg, Q=S("RF = AddFlags[RA][%s];\n  RA=RA+%s", RS, RS), Q)
#define M_ADD(Reg) M_ADD_S(#Reg)

#define M_ADC_S(Reg) (RS=Reg, Q=S("i=RA+%s+(RF&1);\n  RF=ZSTable[i&255]|((i&256)>>8)|((RA^i^%s)&H_FLAG)|(((%s^RA^0x80)&(%s^i)&0x80)>>5);\n  RA=i", RS, RS, RS, RS), Q)
#define M_ADC(Reg) M_ADC_S(#Reg)
// to do, if safe.  Not sure if table lookup needs carry bit as input: #define M_ADC(Reg) S("i=RA+%s+(RF&1);\n  RF = AddFlags[RA][%s];\n  RA=i", #Reg, #Reg)

#define M_SUB_S(Reg) (RS=Reg, Q=S("i=RA-%s; RF=ZSTable[i&255]|((i&256)>>8)|N_FLAG|((RA^i^%s)&H_FLAG)|(((%s^RA)&(%s^i)&0x80)>>5); RA=i", RS, RS, RS, RS), Q)
#define M_SUB(Reg)   M_SUB_S(#Reg)

#define M_SBC_S(Reg)  (RS=Reg, Q=S("i=RA-%s-(RF&1); RF=ZSTable[i&255]|((i&256)>>8)|N_FLAG|((RA^i^%s)&H_FLAG)|(((%s^RA)&(%s^i)&0x80)>>5); RA=i", RS, RS, RS, RS), Q)
#define M_SBC(Reg)    M_SBC_S(#Reg)

#define M_CP_S(Reg)   (RS=Reg, Q=S("i=RA-%s; RF=ZSTable[i&255]|((i&256)>>8)|N_FLAG|((RA^i^%s)&H_FLAG)|(((%s^RA)&(%s^i)&0x80)>>5)", RS, RS, RS, RS), Q)
#define M_CP(Reg)     M_CP_S(#Reg)

// todo: check the logic of this: what is the 0x1000 for?  Should it be 0x10000?
#define M_ADDW_S(Reg1,Reg2) (RS=Reg1, RS2=Reg2, Q=S("i=%s+%s; RF=(RF&(S_FLAG|Z_FLAG|V_FLAG))|(((%s^i^%s)&0x1000)>>8)|((i>>16)&1); %s=i", RS, RS2, RS, RS2, RS), Q)
#define M_ADDW(Reg1, Reg2) M_ADDW_S(#Reg1, #Reg2)

#define M_ADCW_S(Reg) (RS=Reg, Q=S("i=HL+%s+(AF&1);RF=(((HL^i^%s)&0x1000)>>8)|((i>>16)&1)|((i&0x8000)>>8)|((i&65535)?0:Z_FLAG)|(((%s^HL^0x8000)&(%s^i)&0x8000)>>13); HL=i", RS, RS, RS, RS), Q)
#define M_ADCW(Reg) M_ADCW_S(#Reg)

#define M_SBCW_S(Reg) (RS=Reg, Q=S("i=HL-%s-(AF&1);RF=(((HL^i^%s)&0x1000)>>8)|((i>>16)&1)|((i&0x8000)>>8)|((i&65535)?0:Z_FLAG)|(((%s^HL)&(%s^i)&0x8000)>>13)|N_FLAG; HL=i", RS, RS, RS, RS), Q)
#define M_SBCW(Reg) M_SBCW_S(#Reg)
